/*
 * main.cpp - import cpe dictionary data to clickhouse database
 *
 * Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>

#include <fstream>
#include <set>
#include <stdexcept>
#include <string>
#include <tuple>
#include <utility>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/xml_parser.hpp>

#include <clickhouse-cpp/client.h>

#include "cve-db-lock.h"
#include "helpers.h"
#include "update.h"
#include "util.h"

void process_cpe_match(const boost::property_tree::ptree &parent, std::set<std::tuple<std::string, std::string, std::string, std::string> > &result)
{
        for (auto iter = parent.begin(); iter != parent.end(); ++iter) {
                auto item_vulnerable = iter->second.get_child_optional("vulnerable");
                auto item_cpe23uri = iter->second.get_child_optional("cpe23Uri");
                if ((!item_vulnerable) || (!item_cpe23uri) || (!item_vulnerable->get_value<bool>(false))) {
                        continue;
                }

                std::string cpe = item_cpe23uri->get_value<std::string>();

                std::vector<std::string> cpe_array;
                split_cpe(cpe, cpe_array);

                if (cpe_array.size() < 13) {
                        continue;
                }

                // cpe_array[3] = vendor
                // cpe_array[4] = product
                // cpe_array[10] = software platform
                std::string software_platform = cpe_array[10];
                if (software_platform == "*")
                {
                    software_platform.clear();
                }

                result.insert(std::move(std::make_tuple(std::move(cpe), std::move(cpe_array[3]), std::move(cpe_array[4]), std::move(software_platform))));
        }
}

void process_nodes(const boost::property_tree::ptree &parent, std::set<std::tuple<std::string, std::string, std::string, std::string> > &result)
{
        for (auto iter = parent.begin(); iter != parent.end(); ++iter) {
                // ignore 'operator' and 'negate' fields for now
                auto cpe_match_item = iter->second.get_child_optional("cpe_match");
                if (cpe_match_item) {
                        process_cpe_match(*cpe_match_item, result);
                }

                auto children = iter->second.get_child_optional("children");
                if (children) {
                        process_nodes(*children, result);
                }
        }
}
void load_cpe_json(const std::string &fname, std::set<std::tuple<std::string, std::string, std::string, std::string> > &result)
{
        boost::property_tree::ptree json_config;

        std::ifstream stream(fname);
        boost::property_tree::read_json(stream, json_config);

        auto cve_items = json_config.get_child_optional("CVE_Items");
        if (!cve_items) {
                std::stringstream str;
                str << "File " << fname << " doesn't have CVE_Items entry";
                throw std::runtime_error(str.str());
        }

        for (auto cve_items_iter = cve_items->begin(); cve_items_iter != cve_items->end(); ++cve_items_iter) {
                // process required 'configurations' field
                auto configurations_object = cve_items_iter->second.get_child_optional("configurations");
                if (!configurations_object) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations' field\n", fname.c_str());
                        continue;
                }

                auto cve_data_version = configurations_object->get_child_optional("CVE_data_version");
                if (!cve_data_version) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations.CVE_data_version' field\n", fname.c_str());
                        continue;
                }

                if (cve_data_version->get_value<std::string>() != "4.0") {
                        fprintf(stderr, "File %s contains CVE entry with unsupported 'CVE_Items.configurations.CVE_data_version' field: expected \"4.0\", got \"%s\"\n", fname.c_str(), cve_data_version->get_value<std::string>().c_str());
                        continue;
                }

                auto nodes_item = configurations_object->get_child_optional("nodes");
                if (!nodes_item) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations.nodes' field\n", fname.c_str());
                        continue;
                }

                process_nodes(*nodes_item, result);
        }
}

static void show_version(void)
{
        const char *msg =
            "\
" PACKAGE " " PACKAGE_VERSION
            "\n\
Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>\n\
" PACKAGE_NAME
            " is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.";
        fprintf(stderr, "%s\n", msg);
}

int main(int argc, char **argv)
{
    try
    {
        bool show_help = false;
        bool _show_version = false;
        bool verbose = false;
        bool skip_download = false;
        std::string nvds;
        std::string _cacert_file;
        uint64_t commit_data_limit = 0;
        std::string clickhouse_config_file;
        std::string cpe_dictionary;

        boost::program_options::options_description desc("General Options");
        desc.add_options()
            ("help,h", boost::program_options::bool_switch(&show_help)->default_value(false), "Show help")
            ("version,v", boost::program_options::bool_switch(&_show_version)->default_value(false), "Show version")
            ("verbose,V", boost::program_options::bool_switch(&verbose)->default_value(false), "Very verbose mode")
            ("skip-download,U", boost::program_options::bool_switch(&skip_download)->default_value(false), "Bypass forced file downloads")
            ("nvd-dir,d", boost::program_options::value< std::string >(&nvds), "NVD directory in filesystem")
            ("cacert,C", boost::program_options::value< std::string >(&_cacert_file), "Path to the combined SSL certificates file (system default is used if not set)")
            ("commit-data-limit,L", boost::program_options::value< uint64_t >(&commit_data_limit), "Commit vulnerabilities data when it reaches specified amount of records")
            ("clickhouse-config,Y", boost::program_options::value< std::string >(&clickhouse_config_file), "Path to file containing configuration for connection to clickhouse database");

        boost::program_options::options_description hidden;
        hidden.add_options()
            ("cpe-dictionary", boost::program_options::value< std::string >(&cpe_dictionary), "CPE dictionary filename");

        boost::program_options::options_description all_options;
        all_options.add(desc);
        all_options.add(hidden);

        boost::program_options::positional_options_description positional_options;
        positional_options.add("cpe-dictionary", 1);

        boost::program_options::variables_map vm;

        boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(positional_options).run(), vm);
        boost::program_options::notify(vm);

        if (show_help) {
                std::stringstream descr;
                descr << all_options;
                printf("Usage: %s [options]\n%s\n", argv[0], descr.str().c_str());
                return EXIT_SUCCESS;
        }

        if (_show_version) {
                show_version();
                return EXIT_SUCCESS;
        }

        if (clickhouse_config_file.empty()) {
                fprintf(stderr, "Clickhouse config is required\n");
                return EXIT_FAILURE;
        }

        std::set<std::tuple<std::string, std::string, std::string, std::string> > old_cpe_records, new_cpe_records;

        std::string db_path = get_db_path(nvds);

        CveDbLock db_lock(db_path);

        bool db_locked = db_lock.cve_db_read_lock(CveDbLock::LOCK_WAIT_SECS);
        if (!db_locked) {
            fputs("Exiting...\n", stderr);
            return EXIT_FAILURE;
        }

        bool update_needed = update_required(db_path);

        if (update_needed && !skip_download) {
            fprintf(stderr, "Update of db forced\n");
        }

        db_lock.cve_db_unlock();
        if (!update_db(db_lock, verbose, skip_download || !update_needed, db_path, _cacert_file, [&new_cpe_records](const std::string &filename) {
                load_cpe_json(filename, new_cpe_records);
            })) {
            fprintf(stderr, "Failed to update database\n");
            return EXIT_FAILURE;
        }

        if (verbose) {
            fprintf(stderr, "Obtaining information about new CPEs\n");
        }

        if (!cpe_dictionary.empty())
        {
            boost::property_tree::ptree xml_config;

            std::ifstream stream(cpe_dictionary);
            boost::property_tree::read_xml(stream, xml_config);

            auto cpe_list = xml_config.get_child_optional("cpe-list");
            if (cpe_list) {
                for (auto child_iter = cpe_list->begin(); child_iter != cpe_list->end(); ++child_iter) {
                    if (child_iter->first != "cpe-item") {
                        continue;
                    }

                    auto cpe23_item = child_iter->second.get_child_optional("cpe-23:cpe23-item");
                    if (cpe23_item) {
                        auto xmlattr = cpe23_item->get_child_optional("<xmlattr>");
                        if (xmlattr) {
                            auto cpe23_item_name = xmlattr->get_child_optional("name");
                            if (cpe23_item_name) {
                                std::string full_cpe = cpe23_item_name->get_value<std::string>();

                                std::vector<std::string> cpe_array;
                                split_cpe(full_cpe, cpe_array);

                                if (cpe_array.size() < 13) {
                                        continue;
                                }

                                // cpe_array[3] = vendor
                                // cpe_array[4] = product
                                // cpe_array[10] = software platform
                                std::string software_platform = cpe_array[10];
                                if (software_platform == "*")
                                {
                                    software_platform.clear();
                                }

                                new_cpe_records.insert(std::move(std::make_tuple(std::move(full_cpe), std::move(cpe_array[3]), std::move(cpe_array[4]), std::move(software_platform))));
                            }
                        }
                    }
                }
            }
        }

        clickhouse::ClientOptions client_options;

        std::string table_name_cpe_dictionary = "CpeDictionary";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(clickhouse_config_file);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_cpe_dictionary, *table_names, "TableNameCpeDictionary", &is_valid_table_name);
            }

            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        std::string query_string;

        /* Create tables if they're missing */
        client.Execute(cpe_dictionary_table_creation_command(table_name_cpe_dictionary));

        if (verbose) {
            fprintf(stderr, "Obtaining information about existing CPEs\n");
        }

        try
        {
            query_string = std::string(
                        "SELECT cd_full_cpe"
                        ", cd_vendor"
                        ", cd_product"
                        ", cd_software_platform"
                        " FROM ("
                            "SELECT cd_full_cpe"
                            ", cd_vendor"
                            ", cd_product"
                            ", cd_software_platform"
                            ", argMax(cd_status, ts) AS cd_status_ts"
                            " FROM ") + table_name_cpe_dictionary + " GROUP BY cd_full_cpe, cd_vendor, cd_product, cd_software_platform"
                        ") WHERE cd_status_ts = 'actual'";

            /* Now fetch all source packages data */
            client.Select(query_string,
                          [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto cd_full_cpe = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    auto cd_vendor = sv_to_string(ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i));
                    auto cd_product = sv_to_string(ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i));
                    auto cd_software_platform = sv_to_string(ensure_not_null(block[3]->As<clickhouse::ColumnString>())->At(i));
                    old_cpe_records.insert(std::move(std::make_tuple(std::move(cd_full_cpe), std::move(cd_vendor), std::move(cd_product), std::move(cd_software_platform))));
                }
            });

            query_string.clear();
        }
        catch (...)
        {
            if (!query_string.empty())
            {
                fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
            }

            throw;
        }

        // remove common entries from both sets
        // after that all that remains in old set is removed records,
        // all that remains in new set - new records
        {
            auto iter1 = old_cpe_records.begin();
            auto iter2 = new_cpe_records.begin();

            while ((iter1 != old_cpe_records.end()) && (iter2 != new_cpe_records.end()))
            {
                if (*iter1 < *iter2)
                {
                    ++iter1;
                }
                else if (*iter1 > *iter2)
                {
                    ++iter2;
                }
                else
                {
                    iter1 = old_cpe_records.erase(iter1);
                    iter2 = new_cpe_records.erase(iter2);
                }
            }
        }

        if (verbose)
        {
            fprintf(stderr, "Removed CPE entries: %zu, new CPE entries: %zu\n", old_cpe_records.size(), new_cpe_records.size());
        }

        auto cd_full_cpe = std::make_shared<clickhouse::ColumnString>();
        auto cd_vendor = std::make_shared<clickhouse::ColumnString>();
        auto cd_product = std::make_shared<clickhouse::ColumnString>();
        auto cd_software_platform = std::make_shared<clickhouse::ColumnString>();
        auto cd_status = std::make_shared<clickhouse::ColumnEnum8>(clickhouse::Type::CreateEnum8({{"actual", 0}, {"removed", 1}}));

        auto commit_data = [&]()
        {
            clickhouse::Block block;

            block.AppendColumn("cd_full_cpe", cd_full_cpe);
            block.AppendColumn("cd_vendor", cd_vendor);
            block.AppendColumn("cd_product", cd_product);
            block.AppendColumn("cd_software_platform", cd_software_platform);
            block.AppendColumn("cd_status", cd_status);

            if (verbose) {
                fprintf(stderr, "Committing data\n");
            }

            client.Insert(table_name_cpe_dictionary, block);

            cd_full_cpe->Clear();
            cd_vendor->Clear();
            cd_product->Clear();
            cd_software_platform->Clear();
            cd_status->Clear();
        };

        for (auto iter = old_cpe_records.begin(); iter != old_cpe_records.end(); ++iter)
        {
            cd_full_cpe->Append(std::get<0>(*iter));
            cd_vendor->Append(std::get<1>(*iter));
            cd_product->Append(std::get<2>(*iter));
            cd_software_platform->Append(std::get<3>(*iter));
            cd_status->Append("removed");

            if ((commit_data_limit) != 0 && (cd_full_cpe->Size() >= commit_data_limit)) {
                commit_data();
            }
        }

        for (auto iter = new_cpe_records.begin(); iter != new_cpe_records.end(); ++iter)
        {
            cd_full_cpe->Append(std::get<0>(*iter));
            cd_vendor->Append(std::get<1>(*iter));
            cd_product->Append(std::get<2>(*iter));
            cd_software_platform->Append(std::get<3>(*iter));
            cd_status->Append("actual");

            if ((commit_data_limit) != 0 && (cd_full_cpe->Size() >= commit_data_limit)) {
                commit_data();
            }
        }

        if (cd_full_cpe->Size() > 0) {
            commit_data();
        }
    }
    catch (const std::exception &e)
    {
        fprintf(stderr, "Error: %s\n", e.what());
        return EXIT_FAILURE;
    }
    catch (...)
    {
        fprintf(stderr, "Unknown failure\n");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
