/*
 * clickhouse-input.cpp - import data from clickhouse database
 *
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "clickhouse-input.h"

#include <stdio.h>

#include <map>
#include <set>
#include <string>
#include <stdexcept>
#include <cctype>
#include <algorithm>
#include <utility>
#include <regex>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/algorithm/string/trim.hpp>

#include <boost/lexical_cast.hpp>

#include <boost/optional.hpp>

#include <clickhouse-cpp/client.h>

#include "helpers.h"
#include "rpm-changelog.h"

namespace {

class VersionInfo
{
public:
    explicit VersionInfo(const std::string &version_string)
    {
        // Split alpha and numeric values, and drop everything after wildcard character
        for (size_t i = 0; i < version_string.size(); )
        {
            if (isalpha(version_string[i]))
            {
                // It's an alphabetic string
                std::string segment;

                for (; isalpha(version_string[i]); ++i)
                {
                    segment += version_string[i];
                }

                m_version.push_back(segment);
            }
            else if (isdigit(version_string[i]))
            {
                // It's a numeric string
                std::string segment;

                for (; isdigit(version_string[i]); ++i)
                {
                    segment += version_string[i];
                }

                m_version.push_back(segment);
            }
            else if (version_string[i] == '*')
            {
                // It's a wildcard, save it for special treatment later

                m_version.push_back("*");
                ++i;
            }
            else if (version_string[i] == '~')
            {
                m_version.push_back("~");
                ++i;
            }
            else
            {
                // Skip unknown characters
                ++i;
            }
        }
    }

    std::string to_string() const
    {
        std::string result;

        if (!m_version.empty())
        {
            result = m_version[0];

            for (size_t i = 1; i < m_version.size(); ++i)
            {
                result += '.';
                result += m_version[i];
            }
        }

        return result;
    }

    bool has_wildcard() const
    {
        for (auto iter = m_version.begin(); iter != m_version.end(); ++iter)
        {
            if (*iter == "*")
            {
                return true;
            }
        }

        return false;
    }

    bool empty() const
    {
        return m_version.empty();
    }

    bool operator<(const VersionInfo &other) const
    {
        size_t min_size = std::min(this->m_version.size(), other.m_version.size());

        for (size_t i = 0; i < min_size; ++i)
        {
            if ((this->m_version[i] == "*") || (other.m_version[i] == "*"))
            {
                return false;
            }

            if ((this->m_version[i] == "~") || (other.m_version[i] == "~"))
            {
                if (this->m_version[i] != "~")
                {
                    return false;
                }

                if (other.m_version[i] != "~")
                {
                    return true;
                }
            }
            else if (isdigit(this->m_version[i][0]))
            {
                if (isdigit(other.m_version[i][0]))
                {
                    auto number_this  = boost::lexical_cast<uint64_t>(this->m_version[i]);
                    auto number_other = boost::lexical_cast<uint64_t>(other.m_version[i]);

                    if (number_this != number_other)
                    {
                        return (number_this < number_other);
                    }
                }
                else
                {
                    // numeric segment is lesser than alpha segment
                    return true;
                }
            }
            else
            {
                if (isalpha(other.m_version[i][0]))
                {
                    auto res = strcmp(this->m_version[i].c_str(), other.m_version[i].c_str());
                    if (res != 0)
                    {
                        return (res < 0);
                    }
                }
                else
                {
                    // alpha segment is greater than numeric segment
                    return false;
                }
            }
        }

        // if all elements until the end of common length match, the one version structure with remaining elements is greater
        // But if next element is a wildcard, it's matching
        return ((this->m_version.size() < other.m_version.size()) && (other.m_version[this->m_version.size()] != "*"));
    }

    bool operator>(const VersionInfo &other) const
    {
        return (other < *this);
    }

    bool operator<=(const VersionInfo &other) const
    {
        return !(*this > other);
    }

    bool operator>=(const VersionInfo &other) const
    {
        return !(*this < other);
    }

    bool operator!=(const VersionInfo &other) const
    {
        return ((*this < other) || (other < *this));
    }

    bool operator==(const VersionInfo &other) const
    {
        return !(*this != other);
    }

private:
    std::vector<std::string> m_version;
};

class VersionAndRelease
{
public:
    VersionAndRelease(const VersionInfo &version, const VersionInfo &release)
        : m_version(version),
        m_release(release)
    {
    }

    const VersionInfo& version() const
    {
        return m_version;
    }

    const VersionInfo& release() const
    {
        return m_release;
    }

    bool operator<(const VersionAndRelease &other) const
    {
        return ((this->m_version < other.m_version)
                || ((this->m_version == other.m_version)
                    && ((!(this->m_version.has_wildcard() && this->m_release.empty()))
                    && (!(other.m_version.has_wildcard() && other.m_release.empty())))
                    && (this->m_release < other.m_release)));
    }

    bool operator>(const VersionAndRelease &other) const
    {
        return (other < *this);
    }

    bool operator<=(const VersionAndRelease &other) const
    {
        return !(*this > other);
    }

    bool operator>=(const VersionAndRelease &other) const
    {
        return !(*this < other);
    }

    bool operator!=(const VersionAndRelease &other) const
    {
        return ((*this < other) || (other < *this));
    }

    bool operator==(const VersionAndRelease &other) const
    {
        return !(*this != other);
    }

private:
    VersionInfo m_version;
    VersionInfo m_release;
};

enum class ComparisonType
{
    GreaterOrEqual = 0,
    Less,
    Greater,
    Equal,
    LessOrEqual,
    NotEqual
};

/* Don't use sorted array due to comparison implementation, and put longer operators first */
const std::vector<std::tuple<std::string, ComparisonType> > comparison_type_strings =
{
    { ">=", ComparisonType::GreaterOrEqual },
    { "<=", ComparisonType::LessOrEqual },
    { "!=", ComparisonType::NotEqual },
    { "==", ComparisonType::Equal },
    { "<",  ComparisonType::Less },
    { ">",  ComparisonType::Greater },
    { "=",  ComparisonType::Equal },
};

bool CheckWithComparisonType(const VersionAndRelease &first, const VersionAndRelease &second, ComparisonType comparsion_type)
{
    switch (comparsion_type)
    {
    case ComparisonType::GreaterOrEqual:
        return (first >= second);

    case ComparisonType::Less:
        return (first < second);

    case ComparisonType::Greater:
        return (first > second);

    case ComparisonType::Equal:
        return (first == second);

    case ComparisonType::LessOrEqual:
        return (first <= second);

    case ComparisonType::NotEqual:
        return (first != second);

    default:
        return false;
    }
}

issue_status checked_vulnerability_enum_convert(int8_t value)
{
    const std::map<int8_t, issue_status> result_map = {
        { 0, issue_status::excluded_as_patched },
        { 1, issue_status::manually_excluded },
    };

    auto iter = result_map.find(value);
    if (iter == result_map.end())
    {
        throw std::runtime_error(std::string("Unrecognised CheckedVulnerabilities cv_status enum value: ") + boost::lexical_cast<std::string>(value));
    }

    return iter->second;
}

} // unnamed namespace

std::list<std::tuple<source_package_base, source_package_input_info> > package_scan_all_packages(const CveCheckTool &self)
{
    std::string query_string;

    try
    {
        clickhouse::ClientOptions client_options;

        std::string table_name_packages = "Packages";
        std::string table_name_changelog = "Changelog";
        std::string table_name_checked_vulnerabilities = "CheckedVulnerabilities";
        std::string table_name_task_iterations = "TaskIterations";
        std::string table_name_task_states = "TaskStates";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(self.clickhouse_config);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_packages, *table_names, "TableNamePackage", &is_valid_table_name);
                optional_set(table_name_changelog, *table_names, "TableNameChangelog", &is_valid_table_name);
                optional_set(table_name_checked_vulnerabilities, *table_names, "TableNameCheckedVulnerabilities", &is_valid_table_name);
                optional_set(table_name_task_iterations, *table_names, "TableNameTaskIterations", &is_valid_table_name);
                optional_set(table_name_task_states, *table_names, "TableNameTaskStates", &is_valid_table_name);
            }

            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadOnly");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        //       pkgname,              vulnerability id,     pkg_evr,                issue status, exclude record hash
        std::map<std::string, std::map<std::string, std::map<std::string, std::tuple<issue_status, uint64_t> > > > fixed_vulnerabilities_map;

        if (!table_name_checked_vulnerabilities.empty())
        {
            query_string = std::string("SELECT vr_id, pkg_name, cv_checked_pkg_evr, cv_id, cv_status FROM ") + table_name_checked_vulnerabilities;

            client.Select(query_string,
                [&] (const clickhouse::Block& block)
                {
                    for (size_t i = 0; i < block.GetRowCount(); ++i)
                    {
                        auto column_pkg_evr = ensure_not_null(ensure_not_null(block[2]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                        auto pkgname = sv_to_string(ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i));
                        auto vr_id = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                        auto checked_vulnerability_hash = ensure_not_null(block[3]->As<clickhouse::ColumnUInt64>())->At(i);
                        auto checked_vulnerability_status = ensure_not_null(block[4]->As<clickhouse::ColumnEnum8>())->At(i);

                        // make vulnerability id uppercase
                        std::transform(vr_id.begin(), vr_id.end(), vr_id.begin(), [](unsigned char c){ return std::toupper(c); });

                        for (size_t j = 0; j < column_pkg_evr->Size(); ++j)
                        {
                            fixed_vulnerabilities_map[pkgname][vr_id][sv_to_string(column_pkg_evr->At(j))] = std::make_tuple<>(checked_vulnerability_enum_convert(checked_vulnerability_status), checked_vulnerability_hash);
                        }
                    }
                }
            );

            query_string.clear();
        }

        if (self.verbose) {
            fprintf(stderr, "Clickhouse input function starting processing package changelogs\n");
        }

        //       changelog entry id, vulnerability id
        std::map<uint64_t, std::set<std::string> > changelog_fixed_vulnerabilities_map;

        query_string = std::string("SELECT chlog_hash, chlog_text FROM ") + table_name_changelog + std::string(" where chlog_text like '%CVE-%'");

        /* fetch all changelogs containing "CVE-" substring */
        client.Select(query_string,
            [&] (const clickhouse::Block& block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    uint64_t package_hash = ensure_not_null(block[0]->As<clickhouse::ColumnUInt64>())->At(i);

                    /* Parse changelog entry */
                    std::vector<std::string> splits;
                    boost::split(splits, ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i), boost::is_any_of("\n"));

                    // rpm_parse_changelog returns fixed vulnerabilities uppercase
                    auto fixed_vulnerabilities = rpm_parse_changelog(splits);
                    changelog_fixed_vulnerabilities_map[package_hash].insert(fixed_vulnerabilities.begin(), fixed_vulnerabilities.end());
                }
            }
        );

        query_string.clear();

        if (self.verbose) {
            fprintf(stderr, "Clickhouse input function starting processing source packages\n");
        }

        std::set<uint64_t> allowed_murmur_hashes;

        if (self.query_type == QueryType::TASK_ID)
        {
            query_string = std::string("WITH '") + self.task_id + std::string("' AS id SELECT titer_srcrpm_hash AS source_hash FROM ")
                + table_name_task_iterations + std::string(" WHERE task_id = id AND task_changed = (SELECT max(task_changed) FROM ")
                + table_name_task_states + std::string(" WHERE task_state IN ('DONE', 'EPERM', 'TESTED') AND task_id = id);");

            client.Select(query_string,
                [&] (const clickhouse::Block &block)
                {
                    for (size_t i = 0; i < block.GetRowCount(); ++i)
                    {
                        allowed_murmur_hashes.insert(ensure_not_null(block[0]->As<clickhouse::ColumnUInt64>())->At(i));
                    }
                }
            );

            query_string.clear();
        }

        std::list<std::tuple<source_package_base, source_package_input_info> > source_packages_list;

        size_t total = 0;
        size_t current = 0;

        query_string = std::string("SELECT pkg_hash, pkg_name, pkg_version, pkg_release, pkg_changelog.hash FROM ")
            + table_name_packages + std::string(" WHERE pkg_sourcepackage = 1");

        if (self.query_type == QueryType::SRC_RPM)
        {
            std::string pkg_version, pkg_release;

            auto delim = self.src_pkg_version.find_first_of("-");
            if ((delim == std::string::npos)
                || (self.src_pkg_version.empty())
                || (delim == 0)
                || (delim == self.src_pkg_version.length() - 1))
            {
                pkg_version = self.src_pkg_version;
            }
            else
            {
                pkg_version = self.src_pkg_version.substr(0, delim);
                pkg_release = self.src_pkg_version.substr(delim + 1);
            }

            query_string += std::string(" and pkg_name = '") + self.src_pkg_name + std::string("' and pkg_version = '") + pkg_version + std::string("'");

            if (!pkg_release.empty())
            {
                query_string += std::string(" and pkg_release = '") + pkg_release + std::string("'");
            }
        }

        /* Now fetch all source packages data */
        client.Select(query_string,
            [&] (const clickhouse::Block& block)
            {
                total += block.GetRowCount();

                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    ++current;

                    source_package_base base;
                    source_package_input_info input_info;

                    base.murmur_hash = ensure_not_null(block[0]->As<clickhouse::ColumnUInt64>())->At(i);
                    base.name = ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i);
                    base.version = ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i);
                    base.release = ensure_not_null(block[3]->As<clickhouse::ColumnString>())->At(i);

                    if (self.query_type == QueryType::TASK_ID)
                    {
                        if (allowed_murmur_hashes.find(base.murmur_hash) == allowed_murmur_hashes.end())
                        {
                            continue;
                        }
                    }

                    std::set<uint64_t> found_changelog_ids;

                    auto changelogs_entries = ensure_not_null(ensure_not_null(block[4]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnUInt64>());

                    for (size_t j = 0; j < changelogs_entries->Size(); ++j)
                    {
                        if (changelog_fixed_vulnerabilities_map.find(changelogs_entries->At(j)) != changelog_fixed_vulnerabilities_map.end())
                        {
                            found_changelog_ids.insert(changelogs_entries->At(j));
                        }
                    }

                    auto fixed_vulnerabilities_iter = fixed_vulnerabilities_map.find(base.name);

                    /* Temporary map copy needed to temporary erase vulnerabilities fixed for this version of package */
                    decltype(fixed_vulnerabilities_map)::mapped_type fixed_vulnerabilities_temp_map;

                    if (fixed_vulnerabilities_iter != fixed_vulnerabilities_map.end())
                    {
                        fixed_vulnerabilities_temp_map = fixed_vulnerabilities_iter->second;
                    }

                    if (!found_changelog_ids.empty())
                    {
                        for (auto found_changelog_ids_iter = found_changelog_ids.begin(); found_changelog_ids_iter != found_changelog_ids.end(); ++found_changelog_ids_iter) {
                            auto changelog_iter = changelog_fixed_vulnerabilities_map.find(*found_changelog_ids_iter);
                            if (changelog_iter != changelog_fixed_vulnerabilities_map.end()) {
                                for (auto iter = changelog_iter->second.begin(); iter != changelog_iter->second.end(); ++iter) {
                                    /*
                                     * If vulnerability fix is mentioned in changelog, just remove it from explicit fixes map,
                                     * but do it only for this temporary copy of map
                                     */
                                    fixed_vulnerabilities_temp_map.erase(*iter);

                                    input_info.vulnerabilities_extra_map[*iter] = vulnerability_extra(issue_status::patched);
                                }
                            }
                        }
                    }

                    if (fixed_vulnerabilities_iter != fixed_vulnerabilities_map.end())
                    {
                        VersionInfo version_info_version(base.version);
                        VersionInfo version_info_release(base.release);

                        for (auto iter = fixed_vulnerabilities_temp_map.begin(); iter != fixed_vulnerabilities_temp_map.end(); ++iter)
                        {
                            for (auto version_iter = iter->second.begin(); version_iter != iter->second.end(); ++version_iter)
                            {
                                auto fixed_version = version_iter->first;

                                ComparisonType comparison_type = ComparisonType::GreaterOrEqual;

                                for (auto comparison_iter = comparison_type_strings.begin(); comparison_iter != comparison_type_strings.end(); ++comparison_iter)
                                {
                                    if (strncmp(std::get<0>(*comparison_iter).c_str(), fixed_version.c_str(), std::get<0>(*comparison_iter).length()) == 0)
                                    {
                                        comparison_type = std::get<1>(*comparison_iter);
                                        fixed_version = (fixed_version.length() > std::get<0>(*comparison_iter).length()) ? boost::trim_copy(fixed_version.substr(std::get<0>(*comparison_iter).length())) : std::string();
                                        break;
                                    }
                                }

                                std::string fixed_release;

                                {
                                    auto delim = fixed_version.find_first_of("-:");
                                    if ((delim != std::string::npos) && (fixed_version[delim] == ':'))
                                    {
                                        if (delim + 1 < fixed_version.length())
                                        {
                                            fixed_version = fixed_version.substr(delim + 1);
                                        }
                                        else
                                        {
                                            fixed_version.clear();
                                        }
                                    }

                                    delim = fixed_version.find_first_of('-');
                                    if (delim != std::string::npos)
                                    {
                                        if (delim + 1 < fixed_version.length())
                                        {
                                            fixed_release = fixed_version.substr(delim + 1);
                                        }

                                        if (delim > 0)
                                        {
                                            fixed_version = fixed_version.substr(0, delim);
                                        }
                                        else
                                        {
                                            fixed_version.clear();
                                        }
                                    }
                                }

                                VersionInfo fixed_version_info_version(fixed_version);
                                VersionInfo fixed_version_info_release(fixed_release);

                                if (CheckWithComparisonType(VersionAndRelease(version_info_version, version_info_release), VersionAndRelease(fixed_version_info_version, fixed_version_info_release), comparison_type))
                                {
                                    std::string vulnerability_name = iter->first;
                                    std::transform(vulnerability_name.begin(), vulnerability_name.end(), vulnerability_name.begin(), [] (unsigned char c) { return std::toupper(c); });

                                    input_info.vulnerabilities_extra_map[vulnerability_name] = vulnerability_extra(std::get<0>(version_iter->second), std::get<1>(version_iter->second));
                                    break;
                                }
                            }
                        }
                    }

                    if (self.verbose) {
                        fprintf(stderr, "Processing source package: %zu/%zu, %s-%s-%s\n", current, total, base.name.c_str(), base.version.c_str(), base.release.c_str());
                    }

                    source_packages_list.emplace_back(std::move(std::make_tuple(std::move(base), std::move(input_info))));
                }
            }
        );

        query_string.clear();

        if (self.verbose) {
            fprintf(stderr, "Clickhouse input function starting processing binary packages\n");
        }

        // ensure to get rid of duplicates
        std::map<uint64_t, std::set<binary_package_t> > binary_packages_map;

        query_string = std::string("SELECT pkg_name, pkg_version, pkg_release, pkg_srcrpm_hash FROM ") + table_name_packages + std::string(" WHERE pkg_sourcepackage = 0");

        /* Now fetch binary packages data */
        client.Select(query_string,
            [&] (const clickhouse::Block &block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    binary_packages_map[ensure_not_null(block[3]->As<clickhouse::ColumnUInt64>())->At(i)].insert(
                        binary_package_t(
                            sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i)),
                            sv_to_string(ensure_not_null(block[1]->As<clickhouse::ColumnString>())->At(i)),
                            sv_to_string(ensure_not_null(block[2]->As<clickhouse::ColumnString>())->At(i))
                        )
                    );
                }
            }
        );

        query_string.clear();

        for (auto iter = source_packages_list.begin(); iter != source_packages_list.end(); ++iter)
        {
            auto hash_iter = binary_packages_map.find(std::get<0>(*iter).murmur_hash);
            if (hash_iter != binary_packages_map.end())
            {
                std::get<1>(*iter).binary_packages = hash_iter->second;
            }
        }

        if (self.verbose) {
            fprintf(stderr, "Clickhouse input function finished processing packages\n");
        }

        return source_packages_list;
    }
    catch (...)
    {
        if (!query_string.empty())
        {
            fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
        }

        throw;
    }
}

void scan_existing_packages_and_vulnerabilities(CveCheckTool &self, const std::list<std::tuple<source_package_base, source_package_input_info> > &packages)
{
    std::string query_string;

    try
    {
        bool only_allowed_hashes = false;
        std::set<uint64_t> allowed_hashes;

        if (self.query_type != QueryType::ALL)
        {
            only_allowed_hashes = true;

            for (auto iter = packages.begin(); iter != packages.end(); ++iter)
            {
                allowed_hashes.insert(std::get<0>(*iter).murmur_hash);
            }
        }

        clickhouse::ClientOptions client_options;

        std::string table_name_vulnerabilities_status = "VulnerabilitiesStatus";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(self.clickhouse_config);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_vulnerabilities_status, *table_names, "TableNameVulnerabilitiesStatus", &is_valid_table_name);
            }

            // since Vulnerabilities table is updated, also use output connection to check current data
            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        client.Execute(vulnerabilities_status_table_creation_command(table_name_vulnerabilities_status));

        query_string = std::string("SELECT vr_id, pkg_hash, argMax(vs_status, ts) FROM ") + table_name_vulnerabilities_status + " GROUP BY pkg_hash, vr_id";

        if (self.verbose) {
            fprintf(stderr, "Obtaining information about existing vulnerabilities\n");
        }

        /* Now fetch all source packages data */
        client.Select(query_string,
            [&] (const clickhouse::Block& block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto vr_id = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));
                    uint64_t murmur_hash = ensure_not_null(block[1]->As<clickhouse::ColumnUInt64>())->At(i);
                    issue_status status = static_cast<issue_status>(ensure_not_null(block[2]->As<clickhouse::ColumnEnum8>())->At(i));

                    if ((!only_allowed_hashes) || (allowed_hashes.find(murmur_hash) != allowed_hashes.end()))
                    {
                        self.existing_vulnerabilities[vr_id][murmur_hash].status = status;
                    }
                }
            }
        );

        query_string.clear();
    }
    catch (...)
    {
        if (!query_string.empty())
        {
            fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
        }

        throw;
    }
}

void scan_package_vendor_product_matching(CveCheckTool &self)
{
    std::string query_string;

    try
    {
        clickhouse::ClientOptions client_options;

        std::string table_name_cpe_dictionary_matches = "CpeDictionaryMatches";

        /* First read config */
        {
            boost::property_tree::ptree clickhouse_config_tree;

            std::ifstream stream(self.clickhouse_config);
            boost::property_tree::read_ini(stream, clickhouse_config_tree);

            auto table_names = clickhouse_config_tree.get_child_optional("Tables");
            if (table_names)
            {
                optional_set(table_name_cpe_dictionary_matches, *table_names, "TableNameCpeDictionaryMatches", &is_valid_table_name);
            }

            // since Vulnerabilities table is updated, also use output connection to check current data
            auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
            if (output_connection)
            {
                fill_connection_options(*output_connection, client_options);
            }
            else
            {
                auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                if (client_connection)
                {
                    fill_connection_options(*client_connection, client_options);
                }
            }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        /* Create tables if they're missing */
        client.Execute(cpe_dictionary_match_table_creation_command(table_name_cpe_dictionary_matches));

        query_string = std::string("SELECT pkg_name"
                                   ", argMax(cdm_match.vendor, ts)"
                                   ", argMax(cdm_match.product, ts)"
                                   ", argMax(cdm_match.software_platform, ts)"
                                   " FROM ") + table_name_cpe_dictionary_matches + std::string(" GROUP BY pkg_name");

        if (self.verbose) {
            fprintf(stderr, "Obtaining information about existing vulnerabilities\n");
        }

        /* Now fetch all source packages data */
        client.Select(query_string,
            [&] (const clickhouse::Block& block)
            {
                for (size_t i = 0; i < block.GetRowCount(); ++i)
                {
                    auto pkg_name = sv_to_string(ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i));

                    auto cdm_vendor = ensure_not_null(ensure_not_null(block[1]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto cdm_product = ensure_not_null(ensure_not_null(block[2]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                    auto cdm_software_platform = ensure_not_null(ensure_not_null(block[3]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                    if ((cdm_vendor->Size() != cdm_product->Size())
                            || (cdm_vendor->Size() != cdm_software_platform->Size()))
                    {
                            std::stringstream str;
                            str << "Columns cdm_match.vendor, cdm_match.product, cdm_match.software_platform have mismatching sizes!";
                            throw std::runtime_error(str.str());
                    }

                    for (size_t j = 0; j < cdm_vendor->Size(); ++j)
                    {
                        self.pkgname_to_cpe_vendor_product_softwareplatform[pkg_name].insert(std::make_tuple(sv_to_string(cdm_vendor->At(j)),
                                                                                                                 sv_to_string(cdm_product->At(j)),
                                                                                                                 sv_to_string(cdm_software_platform->At(j))));
                    }
                }
            }
        );

        query_string.clear();
    }
    catch (...)
    {
        if (!query_string.empty())
        {
            fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
        }

        throw;
    }
}
