/*
 * core.cpp - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "core.h"

#include <fcntl.h>
#include <inttypes.h>
#include <malloc.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <fstream>
#include <map>
#include <memory>
#include <set>
#include <sstream>
#include <stdexcept>
#include <string>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>

#include "helpers.h"
#include "util.h"

std::optional<cve_entry_t> CveDB::get_cve(const std::string &id) const
{
        auto iter = cve_map.find(id);
        if (iter == cve_map.end()) {
                return std::nullopt;
        }

        return iter->second;
}

std::set<issue_t> CveDB::get_issues(const std::string &vendor, const std::string &product, const std::string &software_platform, const std::string &version, std::set<discarded_issue_t> *discarded_issues_set) const
{
        std::set<issue_t> result;

        if (product.empty()) {
                return result;
        }

        auto product_iter = product_issues_map.find(product);
        if (product_iter != product_issues_map.end()) {
                for (auto vendor_iter = product_iter->second.begin(); vendor_iter != product_iter->second.end(); ++vendor_iter) {
                        if (!vendor.empty()) {
                                if (vendor_iter->first != vendor) {
                                        continue;
                                }
                        } else {
                                if (explicit_vendors.find(vendor_iter->first) != explicit_vendors.end()) {
                                        continue;
                                }
                        }

                        for (auto software_platform_iter = vendor_iter->second.begin(); software_platform_iter != vendor_iter->second.end(); ++software_platform_iter)
                        {
                                if ((!software_platform.empty()) && (software_platform_iter->first != software_platform)) {
                                        continue;
                                }

                                for (auto issue_iter = software_platform_iter->second.begin(); issue_iter != software_platform_iter->second.end(); ++issue_iter) {
                                        auto cve_entry_iter = cve_map.find(*issue_iter);
                                        if (cve_entry_iter == cve_map.end()) {
                                                continue;
                                        }

                                        auto cpe_product_iter = cve_entry_iter->second.cpe.find(product);
                                        if (cpe_product_iter == cve_entry_iter->second.cpe.end()) {
                                                continue;
                                        }

                                        auto cpe_vendor_iter = cpe_product_iter->second.find(vendor_iter->first);
                                        if (cpe_vendor_iter == cpe_product_iter->second.end()) {
                                                continue;
                                        }

                                        auto cpe_software_platform_iter = cpe_vendor_iter->second.find(software_platform_iter->first);
                                        if (cpe_software_platform_iter == cpe_vendor_iter->second.end()) {
                                                continue;
                                        }

                                        auto cpe_iter = cpe_software_platform_iter->second.begin();

                                        for ( ; cpe_iter != cpe_software_platform_iter->second.end(); ++cpe_iter) {
                                                if ((cpe_iter->versionStartExcluding) && (strverscmp(version.c_str(), cpe_iter->versionStartExcluding->c_str()) <= 0)) {
                                                        continue;
                                                }

                                                if ((cpe_iter->versionStartIncluding) && (strverscmp(version.c_str(), cpe_iter->versionStartIncluding->c_str()) < 0)) {
                                                        continue;
                                                }

                                                if ((cpe_iter->versionEndExcluding) && (strverscmp(version.c_str(), cpe_iter->versionEndExcluding->c_str()) >= 0)) {
                                                        continue;
                                                }

                                                if ((cpe_iter->versionEndIncluding) && (strverscmp(version.c_str(), cpe_iter->versionEndIncluding->c_str()) > 0)) {
                                                        continue;
                                                }

                                                break;
                                        }

                                        if (cpe_iter != cpe_software_platform_iter->second.end()) {
                                                issue_t item;

                                                item.vulnerability_id = *issue_iter;
                                                item.cpe = *cpe_iter;

                                                result.insert(std::move(item));
                                        } else if (discarded_issues_set != nullptr) {
                                                discarded_issue_t discarded_issue;

                                                discarded_issue.vulnerability_id = *issue_iter;
                                                discarded_issue.discard_reason = discard_reason_version_mismatch;

                                                discarded_issues_set->insert(std::move(discarded_issue));
                                        }
                                }
                        }
                }
        }

        return result;
}

void CveDB::load(const std::string &clickhouse_config_file, bool verbose)
{
        clickhouse::ClientOptions client_options;

        std::string table_name_vulnerabilities = "Vulnerabilities";

        /* First read config */
        {
                boost::property_tree::ptree clickhouse_config_tree;

                std::ifstream stream(clickhouse_config_file);
                boost::property_tree::read_ini(stream, clickhouse_config_tree);

                auto table_names = clickhouse_config_tree.get_child_optional("Tables");
                if (table_names)
                {
                        optional_set(table_name_vulnerabilities, *table_names, "TableNameVulnerabilities", &is_valid_table_name);
                }

                auto output_connection = clickhouse_config_tree.get_child_optional("ConnectionReadWrite");
                if (output_connection)
                {
                        fill_connection_options(*output_connection, client_options);
                }
                else
                {
                        auto client_connection = clickhouse_config_tree.get_child_optional("Connection");
                        if (client_connection)
                        {
                                fill_connection_options(*client_connection, client_options);
                        }
                }
        }

        /* Now connect to clickhouse */
        clickhouse::Client client(client_options);

        std::string query_string;

        /* Create tables if they're missing */
        client.Execute(vulnerabilities_table_creation_command(table_name_vulnerabilities));

        if (verbose) {
                fprintf(stderr, "Obtaining information about existing vulnerabilities\n");
        }

        try
        {
                query_string = std::string(
                            "SELECT vr_id"
                            ", vr_cpe_string"
                            ", vr_cpe_version_start_excluding"
                            ", vr_cpe_version_start_including"
                            ", vr_cpe_version_end_excluding"
                            ", vr_cpe_version_end_including"
                            " FROM ("
                                "SELECT vr_id"
                                ", argMax(vr_cpe.string, ts) as vr_cpe_string"
                                ", argMax(vr_cpe.version_start_excluding, ts) as vr_cpe_version_start_excluding"
                                ", argMax(vr_cpe.version_start_including, ts) as vr_cpe_version_start_including"
                                ", argMax(vr_cpe.version_end_excluding, ts) as vr_cpe_version_end_excluding"
                                ", argMax(vr_cpe.version_end_including, ts) as vr_cpe_version_end_including"
                                ", argMax(vr_status, ts) AS vr_status_ts"
                                " FROM ") + table_name_vulnerabilities + " GROUP BY vr_id"
                            ") WHERE vr_status_ts = 'actual'";

                /* Now fetch all source packages data */
                client.Select(query_string,
                              [&] (const clickhouse::Block &block)
                {
                        for (size_t i = 0; i < block.GetRowCount(); ++i)
                        {
                                cve_entry_t entry;

                                entry.id = ensure_not_null(block[0]->As<clickhouse::ColumnString>())->At(i);

                                auto vr_cpe_strings = ensure_not_null(ensure_not_null(block[1]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                                auto vr_cpe_start_excluding = ensure_not_null(ensure_not_null(block[2]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                                auto vr_cpe_start_including = ensure_not_null(ensure_not_null(block[3]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                                auto vr_cpe_end_excluding = ensure_not_null(ensure_not_null(block[4]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());
                                auto vr_cpe_end_including = ensure_not_null(ensure_not_null(block[5]->As<clickhouse::ColumnArray>())->GetAsColumn(i)->As<clickhouse::ColumnString>());

                                if ((vr_cpe_strings->Size() != vr_cpe_start_excluding->Size())
                                        || (vr_cpe_strings->Size() != vr_cpe_start_including->Size())
                                        || (vr_cpe_strings->Size() != vr_cpe_end_excluding->Size())
                                        || (vr_cpe_strings->Size() != vr_cpe_end_including->Size()))
                                {
                                        std::stringstream str;
                                        str << "Columns vr_cpe.string, vr_cpe.version_start_excluding, vr_cpe.version_start_including, vr_cpe.version_end_excluding, vr_cpe.version_end_including have mismatching sizes!";
                                        throw std::runtime_error(str.str());
                                }

                                for (size_t j = 0; j < vr_cpe_strings->Size(); ++j)
                                {
                                        cpe_info info;
                                        info.cpe = vr_cpe_strings->At(j);

                                        std::vector<std::string> cpe_array;
                                        split_cpe(info.cpe, cpe_array);

                                        if (cpe_array.size() < 13) {
                                            continue;
                                        }

                                        auto tmp_value = vr_cpe_start_excluding->At(j);
                                        if (!tmp_value.empty())
                                        {
                                                info.versionStartExcluding = tmp_value;
                                        }

                                        tmp_value = vr_cpe_start_including->At(j);
                                        if (!tmp_value.empty())
                                        {
                                                info.versionStartIncluding = tmp_value;
                                        }

                                        tmp_value = vr_cpe_end_excluding->At(j);
                                        if (!tmp_value.empty())
                                        {
                                                info.versionEndExcluding = tmp_value;
                                        }

                                        tmp_value = vr_cpe_end_including->At(j);
                                        if (!tmp_value.empty())
                                        {
                                                info.versionEndIncluding = tmp_value;
                                        }

                                        // cpe_array[4] = product
                                        // cpe_array[3] = vendor
                                        // cpe_array[10] = software platform
                                        std::string software_platform = cpe_array[10];
                                        if (software_platform == "*")
                                        {
                                            software_platform.clear();
                                        }

                                        product_issues_map[cpe_array[4]][cpe_array[3]][software_platform].insert(entry.id);
                                        entry.cpe[cpe_array[4]][cpe_array[3]][software_platform].insert(std::move(info));
                                }

                                std::string vr_id = entry.id;

                                cve_map[vr_id] = std::move(entry);
                            }
                    });

                query_string.clear();
        }
        catch (...)
        {
                if (!query_string.empty())
                {
                        fprintf(stderr, "Clickhouse query: %s\n", query_string.c_str());
                }

                throw;
        }
}
