/*
 * rpm-changelog.cpp - rpm changelog parsing
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "rpm-changelog.h"

#include <algorithm>
#include <cctype>
#include <regex>

#include <boost/algorithm/string.hpp>

static const std::regex cve_regex_fixes_start("^(- )fixe[ds][:[:space:]]*$", std::regex_constants::ECMAScript | std::regex_constants::icase);
static const std::regex cve_regex_opt_start("^[[:space:]]+[[:punct:][:space:]]*([[:alnum:][:punct:]]+)[[:print:]]*", std::regex_constants::ECMAScript | std::regex_constants::icase);
static const std::regex cve_regex_colon(":$", std::regex_constants::ECMAScript | std::regex_constants::icase);
static const std::regex cve_regex_cve("^CVE-[[:digit:]]{4}-[[:digit:]]{4,}$", std::regex_constants::ECMAScript | std::regex_constants::icase);
static const std::regex cve_regex_fixes_short("\\((by[[:space:]]*|fixe[ds]):?", std::regex_constants::ECMAScript | std::regex_constants::icase);

static inline void parse_array(const std::vector<std::string> &array, std::set<std::string> &result) {
        for (auto iter = array.begin(); iter != array.end(); ++iter) {
                if (std::regex_match(*iter, cve_regex_cve)) {
                        std::string cve = *iter;
                        std::transform(cve.begin(), cve.end(), cve.begin(), [](unsigned char c){ return std::toupper(c); });
                        result.insert(cve);
                }
        }
}

static inline void parse_string(std::string str, std::set<std::string> &result) {
        for (auto pos = str.find(','); pos != std::string::npos; pos = str.find(',', pos))
        {
            str.replace(pos, 1, " ");
        }
    
        for (auto pos = str.find("and"); pos != std::string::npos; pos = str.find("and", pos))
        {
            str.replace(pos, sizeof("and") - 1, "   ");
        }

        std::vector<std::string> splits;
        boost::split(splits, str, boost::is_any_of(" "));

        parse_array(splits, result);
}

static inline void scan_short_form(const std::vector<std::string> &lines, std::vector<std::string>::const_iterator &line_iter, std::set<std::string> &result) {
        std::smatch match_info;

        if (std::regex_search(*line_iter, match_info, cve_regex_fixes_short)) {
                std::string str = match_info.suffix();

                auto last_bracket = str.find(')');

                while (last_bracket == std::string::npos) {
                        if (++line_iter == lines.end()) {
                                return;
                        }

                        str.append(" ");
                        str.append(*line_iter);

                        last_bracket = str.find(')');
                }

                str = str.substr(0, last_bracket);

                parse_string(str, result);
        }
}

std::set<std::string> rpm_parse_changelog(const std::vector<std::string> &lines) {
        std::set<std::string> result;

        for (auto line_iter = lines.begin(); line_iter != lines.end(); ++line_iter) {
                if (std::regex_match(*line_iter, cve_regex_fixes_start)) {
                        if (++line_iter == lines.end()) {
                                return result;
                        }

                        std::smatch match_info;

                        while (std::regex_match(*line_iter, match_info, cve_regex_opt_start)) {
                                if (match_info.size() >= 2) {
                                        std::string start = std::regex_replace(static_cast<const std::string &>(match_info[1]), cve_regex_colon, std::string());

                                        parse_string(start, result);
                                }

                                if (++line_iter == lines.end()) {
                                        return result;
                                }
                        }
                }

                scan_short_form(lines, line_iter, result);
                if (line_iter == lines.end()) {
                        return result;
                }
        }

        return result;
}
