/*
 * parse-json.cpp - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 * Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "parse-json.h"

#include "util.h"

#include <cmath>
#include <utility>

#include <boost/algorithm/string.hpp>

#include <boost/lexical_cast.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#define NVD_CVE_URI "https://nvd.nist.gov/vuln/detail/"

namespace {

void process_cpe_match(const boost::property_tree::ptree &parent, std::set<cpe_info> &cpe_set)
{
        for (auto iter = parent.begin(); iter != parent.end(); ++iter) {
                auto item_vulnerable = iter->second.get_child_optional("vulnerable");
                auto item_cpe23uri = iter->second.get_child_optional("cpe23Uri");
                if ((!item_vulnerable) || (!item_cpe23uri) || (!item_vulnerable->get_value<bool>(false))) {
                        continue;
                }

                auto item_version_start_excluding = iter->second.get_child_optional("versionStartExcluding");
                auto item_version_start_including = iter->second.get_child_optional("versionStartIncluding");
                auto item_version_end_excluding = iter->second.get_child_optional("versionEndExcluding");
                auto item_version_end_including = iter->second.get_child_optional("versionEndIncluding");

                if (item_version_start_excluding && item_version_start_including) {
                        continue;
                }

                if (item_version_end_excluding && item_version_end_including) {
                        continue;
                }

                cpe_info info_item;
                info_item.cpe = item_cpe23uri->get_value<std::string>();

                std::vector<std::string> cpe_array;
                split_cpe(info_item.cpe, cpe_array);

                if (cpe_array.size() < 13) {
                        continue;
                }

                if (item_version_start_excluding) {
                        auto tmp_value = item_version_start_excluding->get_value<std::string>();
                        if (!tmp_value.empty()) {
                                info_item.versionStartExcluding = tmp_value;
                        }
                }

                if (item_version_start_including) {
                        auto tmp_value = item_version_start_including->get_value<std::string>();
                        if (!tmp_value.empty()) {
                                info_item.versionStartIncluding = tmp_value;
                        }
                }

                if (item_version_end_excluding) {
                        auto tmp_value = item_version_end_excluding->get_value<std::string>();
                        if (!tmp_value.empty()) {
                                info_item.versionEndExcluding = tmp_value;
                        }
                }

                if (item_version_end_including) {
                        auto tmp_value = item_version_end_including->get_value<std::string>();
                        if (!tmp_value.empty()) {
                                info_item.versionEndIncluding = tmp_value;
                        }
                }

                if ((!item_version_start_excluding) && (!item_version_start_including) && (!item_version_end_excluding) && (!item_version_end_including)) {
                        if ((!cpe_array[5].empty()) && (cpe_array[5] != "*") && (cpe_array[5] != "-")) {
                                info_item.versionStartIncluding = cpe_array[5];
                                info_item.versionEndIncluding = cpe_array[5];
                        }
                }

                cpe_set.insert(std::move(info_item));
        }
}

void process_nodes(const boost::property_tree::ptree &parent, std::set<cpe_info> &cpe_set)
{
        for (auto iter = parent.begin(); iter != parent.end(); ++iter) {
                // ignore 'operator' and 'negate' fields for now
                auto cpe_match_item = iter->second.get_child_optional("cpe_match");
                if (cpe_match_item) {
                        process_cpe_match(*cpe_match_item, cpe_set);
                }

                auto children = iter->second.get_child_optional("children");
                if (children) {
                        process_nodes(*children, cpe_set);
                }
        }
}

} // unnamed namespace

bool vulnerability_record::operator==(const vulnerability_record &other) const
{
    return (description == other.description)
        && (url == other.url)
        && (std::nextafter(score, std::numeric_limits<decltype(score)>::lowest()) <= other.score)
        && (std::nextafter(score, std::numeric_limits<decltype(score)>::max()) >= other.score)
        && (attacktype == other.attacktype)
        && (uris == other.uris)
        && (modifieddate == other.modifieddate)
        && (cpe == other.cpe)
        && (configurations == other.configurations)
        && (type == other.type)
        && (status == other.status);
}

/**
 * Parse an NVD json database
 *
 * @param fname Path to the nvd db
 */
void load_vulnerabilities_json(const std::string &fname, std::map<std::string, vulnerability_record> &vulnerabilities)
{
        boost::property_tree::ptree json_config;

        std::ifstream stream(fname);
        boost::property_tree::read_json(stream, json_config);

        auto cve_items = json_config.get_child_optional("CVE_Items");
        if (!cve_items) {
                std::stringstream str;
                str << "File " << fname << " doesn't have CVE_Items entry";
                throw std::runtime_error(str.str());
        }

        for (auto cve_items_iter = cve_items->begin(); cve_items_iter != cve_items->end(); ++cve_items_iter) {
                vulnerability_record record;

                // process required 'cve' field
                auto cve_object = cve_items_iter->second.get_child_optional("cve");
                if (!cve_object) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.cve' field\n", fname.c_str());
                        continue;
                }

                auto cve_metadata = cve_object->get_child_optional("CVE_data_meta");
                if (!cve_metadata) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.cve.CVE_data_meta' field\n", fname.c_str());
                        continue;
                }

                auto cve_id = cve_metadata->get_child_optional("ID");
                if (!cve_id) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.cve.CVE_data_meta.ID' field\n", fname.c_str());
                        continue;
                }

                std::string vr_id = boost::to_upper_copy<std::string>(cve_id->get_value<std::string>());

                auto cve_description = cve_object->get_child_optional("description");
                if (!cve_description) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.description' field for cve %s\n", fname.c_str(), vr_id.c_str());
                        continue;
                }

                auto cve_description_data = cve_description->get_child_optional("description_data");
                if (!cve_description_data) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.description.description_data' field for cve %s\n", fname.c_str(), vr_id.c_str());
                        continue;
                }

                for (auto description_iter = cve_description_data->begin(); description_iter != cve_description_data->end(); ++description_iter) {
                        auto lang_item = description_iter->second.get_child_optional("lang");
                        if (!lang_item) {
                                fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.description.description_data.lang' field for cve %s\n", fname.c_str(), vr_id.c_str());
                                continue;
                        }

                        if (lang_item->get_value<std::string>() != "en") {
                                continue;
                        }

                        auto cve_description_item = description_iter->second.get_child_optional("value");
                        if (!cve_description_item) {
                                fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.description.description_data.value' field for cve %s\n", fname.c_str(), vr_id.c_str());
                                continue;
                        }

                        record.description = cve_description_item->get_value<std::string>();
                }

                // prcess optional 'references' field
                auto references_item = cve_object->get_child_optional("references");
                if (references_item) {
                        auto reference_data = references_item->get_child_optional("reference_data");
                        if (reference_data) {
                                for (auto references_iter = reference_data->begin(); references_iter != reference_data->end(); ++references_iter) {
                                        auto url_item = references_iter->second.get_child_optional("url");
                                        if (url_item) {
                                                record.uris.insert(url_item->get_value<std::string>());
                                        }
                                }
                        }
                }

                // process required 'configurations' field
                auto configurations_object = cve_items_iter->second.get_child_optional("configurations");
                if (!configurations_object) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations' field for cve %s\n", fname.c_str(), vr_id.c_str());
                        continue;
                }

                auto cve_data_version = configurations_object->get_child_optional("CVE_data_version");
                if (!cve_data_version) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations.CVE_data_version' field for cve %s\n", fname.c_str(), vr_id.c_str());
                        continue;
                }

                if (cve_data_version->get_value<std::string>() != "4.0") {
                        fprintf(stderr, "File %s contains CVE entry with unsupported 'CVE_Items.configurations.CVE_data_version' field for cve %s: expected \"4.0\", got \"%s\"\n", fname.c_str(), vr_id.c_str(), cve_data_version->get_value<std::string>().c_str());
                        continue;
                }

                auto nodes_item = configurations_object->get_child_optional("nodes");
                if (!nodes_item) {
                        fprintf(stderr, "File %s contains CVE entry without 'CVE_Items.configurations.nodes' field for cve %s\n", fname.c_str(), vr_id.c_str());
                        continue;
                }

                process_nodes(*nodes_item, record.cpe);

                // process optional 'lastModifiedDate' field
                auto date_object = cve_items_iter->second.get_child_optional("lastModifiedDate");
                if (date_object) {
                        record.modifieddate = parse_date(date_object->get_value<std::string>());
                }

                // process optional 'impact' field
                auto impact_object = cve_items_iter->second.get_child_optional("impact");
                if (impact_object) {
                        auto base_metric_v3 = impact_object->get_child_optional("baseMetricV3");
                        if (base_metric_v3) {
                                auto cvss_item = base_metric_v3->get_child_optional("cvssV3");
                                if (cvss_item) {
                                        auto score_item = cvss_item->get_child_optional("baseScore");
                                        if (score_item) {
                                                try
                                                {
                                                        record.score = boost::lexical_cast<double>(score_item->get_value<std::string>());
                                                } catch (...)
                                                {
                                                        // ignore
                                                }
                                        }

                                        auto attack_vector_item = cvss_item->get_child_optional("attackVector");
                                        if (attack_vector_item) {
                                                record.attacktype = attack_vector_item->get_value<std::string>();
                                        }
                                }
                        } else {
                                auto base_metric_v2 = impact_object->get_child_optional("baseMetricV2");
                                if (base_metric_v2) {
                                        auto cvss_item = base_metric_v2->get_child_optional("cvssV2");
                                        if (cvss_item) {
                                                auto score_item = cvss_item->get_child_optional("baseScore");
                                                if (score_item) {
                                                        try
                                                        {
                                                                record.score = boost::lexical_cast<double>(score_item->get_value<std::string>());
                                                        } catch (...)
                                                        {
                                                                // ignore
                                                        }
                                                }

                                                auto attack_vector_item = cvss_item->get_child_optional("accessVector");
                                                if (attack_vector_item) {
                                                        record.attacktype = attack_vector_item->get_value<std::string>();
                                                }
                                        }
                                }
                        }
                }

                {
                    std::stringstream configurations_stream;
                    boost::property_tree::write_json(configurations_stream, *configurations_object);

                    record.configurations = configurations_stream.str();
                }

                record.url = std::string(NVD_CVE_URI) + vr_id;
                record.type = "CVE";

                vulnerabilities[vr_id] = std::move(record);
        }
}
