/*
 * fetch.c - cve-check-tool helpers
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <curl/curl.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>

#include <fstream>
#include <memory>

#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>

#include "fetch.h"
#include "util.h"

struct fetch_t {
        std::unique_ptr<FILE, void(*)(FILE*)> f;
        std::string target;

        fetch_t()
            : f(nullptr, [](FILE *item) {
                    if (item != nullptr) {
                            fclose(item);
                    }
            })
        {
        }
};

static size_t write_func(char *ptr, size_t size, size_t nmemb, void *outstream)
{
        struct fetch_t *f = (struct fetch_t*) outstream;

        if (!f->f) {
                f->f.reset(fopen(f->target.c_str(), "wb"));
        }

        if (!f->f) {
                return -1;
        }

        return fwrite(ptr, size, nmemb, f->f.get());
}

FetchStatus fetch_uri(const std::string &uri, const std::string &target, bool verbose, const std::string &cacert_file)
{
        CURLcode res;
        struct stat st;

        if (verbose) {
                fprintf(stderr, "Downloading \"%s\" to \"%s\"\n", uri.c_str(), target.c_str());
        }

        std::unique_ptr<CURL, void(*)(CURL*)> curl(curl_easy_init(), [](CURL *item) {
                if (item != nullptr) {
                        curl_easy_cleanup(item);
                }
        });

        if (!curl) {
                return FETCH_STATUS_FAIL;
        }

        if (!cacert_file.empty()) {
                res = curl_easy_setopt(curl.get(), CURLOPT_CAINFO, cacert_file.c_str());
                if (res != CURLE_OK) {
                        return FETCH_STATUS_FAIL;
                }
        }

        if (stat(target.c_str(), &st) == 0) {
                res = curl_easy_setopt(curl.get(), CURLOPT_TIMECONDITION, CURL_TIMECOND_IFMODSINCE);
                if (res != CURLE_OK) {
                        return FETCH_STATUS_FAIL;
                }

                res = curl_easy_setopt(curl.get(), CURLOPT_TIMEVALUE, st.st_mtime);
                if (res != CURLE_OK) {
                        return FETCH_STATUS_FAIL;
                }
        }

        res = curl_easy_setopt(curl.get(), CURLOPT_URL, uri.c_str());
        if (res != CURLE_OK) {
                return FETCH_STATUS_FAIL;
        }

        if (verbose) {
                (void) curl_easy_setopt(curl.get(), CURLOPT_NOPROGRESS, 0L);
        }

        res = curl_easy_setopt(curl.get(), CURLOPT_WRITEFUNCTION, write_func);
        if (res != CURLE_OK) {
                return FETCH_STATUS_FAIL;
        }

        fetch_t f;
        f.target = target;

        res = curl_easy_setopt(curl.get(), CURLOPT_WRITEDATA, &f);
        if (res != CURLE_OK) {
                return FETCH_STATUS_FAIL;
        }

        /* Handle future redirects from NVD */
        res = curl_easy_setopt(curl.get(), CURLOPT_FOLLOWLOCATION, 1L);
        if (res != CURLE_OK) {
                return FETCH_STATUS_FAIL;
        }

        res = curl_easy_perform(curl.get());
        if (res != CURLE_OK) {
                return FETCH_STATUS_FAIL;
        }

        return (f.f) ? FETCH_STATUS_UPDATE : FETCH_STATUS_OK;
}

void gunzip_file(const std::string &path)
{
        std::string newpath = path;

        if (newpath.ends_with(".gz")) {
                newpath = newpath.substr(0, newpath.length() - sizeof(".gz") + 1);
        }

        std::ifstream infile(path, std::ios_base::in | std::ios_base::binary);
        boost::iostreams::filtering_streambuf<boost::iostreams::input> in;
        in.push(boost::iostreams::gzip_decompressor());
        in.push(infile);

        std::ofstream outfile(newpath, std::ios_base::out | std::ios_base::binary);

        boost::iostreams::copy(in, outfile);
}
