/*
 * helpers.cpp - various helpers for cve-check-tool
 *
 * Copyright (C) 2021 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "helpers.h"

#include <string.h>

#include <algorithm>
#include <cctype>

bool is_valid_table_name(const std::string &value)
{
    return (std::find_if_not(value.begin(), value.end(), [] (const char c) { return (std::isalnum(c) || (strchr("_", c) != NULL)); } ) == value.end());
}

void fill_connection_options(const boost::property_tree::ptree &connection_section, clickhouse::ClientOptions &client_options)
{
    optional_set(client_options.host, connection_section, "Hostname");
    optional_set(client_options.port, connection_section, "Port");
    optional_set(client_options.default_database, connection_section, "DefaultDatabase");
    optional_set(client_options.user, connection_section, "Username");
    optional_set(client_options.password, connection_section, "Password");
    optional_set(client_options.ping_before_query, connection_section, "PingBeforeQuery");
    optional_set(client_options.send_retries, connection_section, "SendRetries");

    auto retry_timeout = connection_section.get_child_optional("RetryTimeout");
    if (retry_timeout)
    {
        client_options.retry_timeout = std::chrono::seconds(retry_timeout->get_value<int>());
    }

    auto compression_method = connection_section.get_child_optional("CompressionMethod");
    if (compression_method)
    {
        const auto value = compression_method->get_value<std::string>();

        const std::map<std::string, clickhouse::CompressionMethod> compression_method_map =
        {
            { "None", clickhouse::CompressionMethod::None },
            { "LZ4", clickhouse::CompressionMethod::LZ4 },
        };

        auto iter = compression_method_map.find(value);
        if (iter != compression_method_map.end())
        {
            client_options.compression_method = iter->second;
        }
        else
        {
            client_options.compression_method = clickhouse::CompressionMethod::None;
        }
    }
}

std::string vulnerabilities_table_creation_command(const std::string &table_name)
{
    std::stringstream str;

    // vr_score could be Decimal32, but it's not supported by clickhouse-cpp yet.
    str << "CREATE TABLE IF NOT EXISTS " << table_name
        << " (vr_id String"
           ", vr_description String"
           ", vr_url String"
           ", vr_score Float64"
           ", vr_attacktype LowCardinality(String)"
           ", vr_uris Array(String)"
           ", vr_modifieddate DateTime"
           ", vr_cpe Nested(string String, version_start_excluding String, version_start_including String, version_end_excluding String, version_end_including String)"
           ", vr_configurations String"
           ", vr_type LowCardinality(String)"
           ", vr_status Enum8('actual' = 0, 'removed' = 1)"
           ", ts DateTime MATERIALIZED now()"
           ") ENGINE = MergeTree ORDER BY (vr_id, vr_modifieddate)";

    return str.str();
}

std::string vulnerabilities_status_table_creation_command(const std::string &table_name)
{
    std::stringstream str;
    str << "CREATE TABLE IF NOT EXISTS " << table_name
        << " (pkg_hash UInt64"
           ", vr_id String"
           ", vs_status Enum8('check' = 0, 'patched' = 1, 'discarded_by_version_check' = 2, 'manually_excluded' = 3, 'removed' = 4)"
           ", vs_cpe String"
           ", vs_cpe_version_start_excluding String"
           ", vs_cpe_version_start_including String"
           ", vs_cpe_version_end_excluding String"
           ", vs_cpe_version_end_including String"
           ", cv_id Nullable(UInt64)"
           ", ts DateTime MATERIALIZED now()"
           ") ENGINE = MergeTree ORDER BY (pkg_hash, vr_id)";

    return str.str();
}

std::string absent_packages_table_creation_command(const std::string &table_name)
{
    std::stringstream str;
    str << "CREATE TABLE IF NOT EXISTS " << table_name
        << " (ap_vendor_name String"
           ", ap_product_name String"
           ", ap_software_platform String"
           ", vr_id String"
           ", ap_status Enum8('actual' = 0, 'removed' = 1)"
           ", ts DateTime MATERIALIZED now()"
           ") ENGINE = MergeTree ORDER BY (ap_vendor_name, ap_product_name, vr_id)";

    return str.str();
}

std::string cpe_dictionary_table_creation_command(const std::string &table_name)
{
    std::stringstream str;
    str << "CREATE TABLE IF NOT EXISTS " << table_name
        << " (cd_full_cpe String"
           ", cd_vendor String"
           ", cd_product String"
           ", cd_software_platform String"
           ", cd_status Enum8('actual' = 0, 'removed' = 1)"
           ", ts DateTime MATERIALIZED now()"
           ") ENGINE = MergeTree ORDER BY (cd_full_cpe)";

    return str.str();
}

std::string cpe_dictionary_match_table_creation_command(const std::string &table_name)
{
    std::stringstream str;
    str << "CREATE TABLE IF NOT EXISTS " << table_name
        << " (pkg_name String"
           ", cdm_match Nested(vendor String, product String, software_platform String)"
           ", ts DateTime MATERIALIZED now()"
           ") ENGINE = MergeTree ORDER BY (pkg_name)";

    return str.str();
}

std::string sv_to_string( std::string_view s )
{
    return std::string(s.data(), s.size());
}
