/*
 * update.c - cve-check-tool
 *
 * Copyright (C) 2015 Intel Corporation
 *
 * cve-check-tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>

#include <memory>
#include <sstream>
#include <stdexcept>

#include <curl/curl.h>
#include <openssl/sha.h>

#include <boost/scope_exit.hpp>

#include "cve-db-lock.h"
#include "util.h"

#include "update.h"
#include "fetch.h"

#define YEAR_START 2002
#define URI_PREFIX "https://nvd.nist.gov/feeds/json/cve/1.1"
#define NVD_DIR "NVDS"

#define UPDATE_THRESHOLD 7200
#define UPDATE_DB_MARKER_SUFFIX "cve.update_db"

static std::string get_home_dir()
{
        std::string home;

        home = getenv("HOME");
        if (home.empty()) {
                struct passwd *p;

                p = getpwuid(getuid());
                if ((!p) || (!p->pw_dir)) {
                        throw std::runtime_error("Failed to determine home directory");
                }

                home = p->pw_dir;
        }

        return home;
}

std::string get_db_path(const std::string &path)
{
        std::string dir;

        if (path.empty()) {
                dir = get_home_dir();
                dir.append("/" NVD_DIR);
        } else {
                dir = path;
        }

        if (mkdir(dir.c_str(), S_IRWXU | S_IRWXG | S_IRWXO)) {
                if (errno != EEXIST) {
                        std::stringstream str;
                        str << "Failed to create directory " << dir << ", error: " << errno << ", " << strerror(errno);
                        throw std::runtime_error(str.str());
                }

                if (!cve_is_dir(dir)) {
                    std::stringstream str;
                    str << dir << " is not a directory";
                    throw std::runtime_error(str.str());
                }
        }

        return dir;
}

static std::string nvdcve_meta_get_val(FILE *f, const std::string &field)
{
        for (;;) {
                char field_name[256], field_value[256];
                int ret;

                ret = fscanf(f, " %255[^: \f\n\r\t\v] :%255s", field_name, field_value);
                if (ret != 2) {
                        if (ret != EOF) {
                                continue;
                        }
                        if (ferror(f)) {
                                if (errno == EINTR || errno == EAGAIN) {
                                        clearerr(f);
                                        continue;
                                }
                        }
                        return std::string();
                }
                if (strcmp(field_name, field.c_str()) == 0) {
                        return field_value;
                }
        }
}

static bool nvdcve_data_ok(const std::string &meta, const std::string &data)
{
        std::string csum_meta;
        char csum_data[SHA256_DIGEST_LENGTH * 2 + 1];
        unsigned char digest[SHA256_DIGEST_LENGTH];
        struct stat st;
        int fdata;
        void *buffer;
        size_t length;

        /* Get digest from the NVD META file */
        std::unique_ptr<FILE, void(*)(FILE*)> fmeta(fopen(meta.c_str(), "r"), [](FILE *item) {
                if (item != nullptr) {
                        fclose(item);
                }
        });

        if (!fmeta) {
                return false;
        }

        csum_meta = nvdcve_meta_get_val(fmeta.get(), "sha256");
        if (csum_meta.empty()) {
                return false;
        }

        /* Compute digest on NVD JSON file */
        fdata = open(data.c_str(), O_RDONLY);
        if (fdata < 0) {
                return false;
        }

        BOOST_SCOPE_EXIT(fdata) {
                close(fdata);
        } BOOST_SCOPE_EXIT_END;

        memset(&st, 0, sizeof(st));
        if (fstat(fdata, &st)) {
                return false;
        }
        length = st.st_size;

        buffer = mmap(NULL, length, PROT_READ, MAP_PRIVATE, fdata, 0);
        if (!buffer) {
                return false;
        }

        BOOST_SCOPE_EXIT(buffer, length) {
                munmap(buffer, length);
        } BOOST_SCOPE_EXIT_END;

        if (!SHA256((const unsigned char*) buffer, length, digest)) {
                return false;
        }

        for (size_t i = 0; i < sizeof(digest); i++) {
                size_t idx = i * 2;
                size_t len = sizeof(csum_data) - idx;

                snprintf(&csum_data[idx], len, "%02hhx", digest[i]);
        }

        return (strcasecmp(csum_meta.c_str(), csum_data) == 0);
}

static bool __update_required(const std::string &update_fname)
{
        struct stat st;
        time_t t;

        memset(&st, 0, sizeof(st));
        if (stat(update_fname.c_str(), &st) != 0) {
                return true;
        }

        t = time(NULL);
        if (difftime(t, st.st_mtime) >= UPDATE_THRESHOLD) {
                return true;
        }

        return false;
}

bool update_required(const std::string &db_path)
{
        std::string u_fname = make_db_dot_fname(db_path, UPDATE_DB_MARKER_SUFFIX);
        return __update_required(u_fname);
}

#ifndef O_NOFOLLOW
#define O_NOFOLLOW 0
#endif

static inline int update_begin(const std::string &update_fname)
{
        const int flags = O_RDONLY | O_CREAT | O_NONBLOCK | O_NOFOLLOW;
        const mode_t mode = S_IRUSR | S_IWUSR;

        return open(update_fname.c_str(), flags, mode);
}

static inline void update_end(int fd)
{
        close(fd);
}

static int do_fetch_update(int year, const std::string &db_dir, bool verbose,
                           const std::string &cacert_file, bool skip_download,
                           const std::function<void(const std::string &)> &callback)
{
        const char nvd_uri[] = URI_PREFIX;
        std::string uri_meta;
        std::string uri_data_gz;
        std::string nvdcve_meta;
        std::string nvdcve_data;
        std::string nvdcve_data_gz;
        std::string nvd_json;
        std::string nvd_json_gz;
        std::string nvd_meta;
        FetchStatus st;
        bool update;

        /* Prepare NVD META file/uri pathes */
        nvd_meta = nvdcve_make_fname(year, "meta");

        nvdcve_meta.reserve(db_dir.length() + sizeof("/") - 1 + nvd_meta.length());
        nvdcve_meta.append(db_dir);
        nvdcve_meta.append("/");
        nvdcve_meta.append(nvd_meta);

        uri_meta.reserve(sizeof(nvd_uri) - 1 + sizeof("/") - 1 + nvd_meta.length());
        uri_meta.append(nvd_uri);
        uri_meta.append("/");
        uri_meta.append(nvd_meta);

        /* Prepare NVD JSON file/uri pathes */
        nvd_json = nvdcve_make_fname(year, "json");

        nvdcve_data.reserve(db_dir.length() + sizeof("/") - 1 + nvd_json.length());
        nvdcve_data.append(db_dir);
        nvdcve_data.append("/");
        nvdcve_data.append(nvd_json);

        nvd_json_gz.reserve(nvd_json.length() + sizeof(".gz") - 1);
        nvd_json_gz.append(nvd_json);
        nvd_json_gz.append(".gz");

        nvdcve_data_gz.reserve(db_dir.length() + sizeof("/") - 1 + nvd_json_gz.length());
        nvdcve_data_gz.append(db_dir);
        nvdcve_data_gz.append("/");
        nvdcve_data_gz.append(nvd_json_gz);

        uri_data_gz.reserve(sizeof(nvd_uri) - 1 + sizeof("/") - 1 + nvd_json_gz.length());
        uri_data_gz.append(nvd_uri);
        uri_data_gz.append("/");
        uri_data_gz.append(nvd_json_gz);

        if (!skip_download) {
                for (int attempt = 0; attempt < 2; ++attempt) {
                        if (attempt > 0) {
                                unlink(nvdcve_meta.c_str());
                                unlink(nvdcve_data.c_str());
                                unlink(nvdcve_data_gz.c_str());
                        }

                        /* Fetch NVD META file */
                        st = fetch_uri(uri_meta, nvdcve_meta, verbose, cacert_file);
                        if (st == FETCH_STATUS_FAIL) {
                                fprintf(stderr, "Failed to fetch %s\n", uri_meta.c_str());
                                return -1;
                        }

                        /* Fetch NVD JSON file */
                        st = fetch_uri(uri_data_gz, nvdcve_data_gz, verbose, cacert_file);
                        switch (st) {
                        case FETCH_STATUS_FAIL:
                                fprintf(stderr, "Failed to fetch %s\n", uri_data_gz.c_str());
                                return -1;
                        case FETCH_STATUS_UPDATE:
                                update = true;
                                break;
                        default:
                                update = !nvdcve_data_ok(nvdcve_meta, nvdcve_data);
                                break;
                        }

                        if (update) {
                                bool ok = false;

                                try
                                {
                                        gunzip_file(nvdcve_data_gz);
                                        ok = nvdcve_data_ok(nvdcve_meta, nvdcve_data);
                                }
                                catch (const std::exception &e)
                                {
                                        // if exception is caught, ok is 'false', try refetching or return error as usual
                                        fprintf(stderr, "Caught exception while unpacking and checking %s: %s\n", nvdcve_data.c_str(), e.what());
                                }
                                catch (...)
                                {
                                        // if exception is caught, ok is 'false', try refetching or return error as usual
                                        fprintf(stderr, "Caught unknown exception while unpacking and checking %s\n", nvdcve_data.c_str());
                                }

                                if (!ok) {
                                        if (attempt == 0) {
                                                continue;
                                        }

                                        fprintf(stderr, "Unpacked data %s is not consistent\n", nvdcve_data.c_str());
                                        /* If CVE_SKIP_VERIFY is set in the environment, don't bail here */
                                        if (!getenv("CVE_SKIP_VERIFY")) {
                                                return -1;
                                        }
                                }
                        }

                        break;
                }
        }

        bool file_exists = cve_file_exists(nvdcve_data);

        if (file_exists) {
                callback(nvdcve_data);
        }

        if (verbose) {
                fprintf(stderr, "%s: %s\n", file_exists ? "Loaded" : "Doesn't exist", nvd_json_gz.c_str());
        }

        return 0;
}

bool update_db(CveDbLock &db_lock, bool verbose, bool skip_download, const std::string &db_path, const std::string &cacert_file, const std::function<void(const std::string &)> &callback)
{
        std::string u_fname;
        struct tm *tm;
        time_t t;
        int year;
        bool db_locked = false;

        t = time(NULL);
        if (t == (time_t)-1) {
                fprintf(stderr, "Can't get local time\n");
                return false;
        }

        tm = localtime(&t);
        if (!tm) {
                fprintf(stderr, "Can't get local time\n");
                return false;
        }

        year = tm->tm_year + 1900;

        u_fname = make_db_dot_fname(db_path, UPDATE_DB_MARKER_SUFFIX);

        db_locked = db_lock.cve_db_write_lock(CveDbLock::LOCK_WAIT_SECS);
        if (!db_locked) {
                fprintf(stderr, "Exiting...\n");
                return false;
        }

        BOOST_SCOPE_EXIT(&db_lock) {
                db_lock.cve_db_unlock();
        } BOOST_SCOPE_EXIT_END;

        if (!skip_download) {
                int u_handle = update_begin(u_fname);
                if (u_handle < 0) {
                        fprintf(stderr, "Can't create timestamp file %s\n", u_fname.c_str());
                        return false;
                }

                update_end(u_handle);
        }

        for (int i = YEAR_START; i <= year + 2; i++) {
                int y = i > year ? year - i : i;
                int rc;

                rc = do_fetch_update(y, db_path, verbose, cacert_file, skip_download, callback);
                if (rc != 0) {
                        return false;
                }
        }

        if (!skip_download) {
                /* Make sure we always update access and modify time on
                 * database file, even if no new data loaded.
                 */
                if (utime(u_fname.c_str(), NULL)) {
                        fprintf(stderr, "Unable to update file access and modify time\n");
                        return false;
                }
        }

        return true;
}
